/* $NetBSD$ */

#include <dev/drm/drmP.h>

int
drm_dmamem_xalloc(struct drm_device *dev, bus_size_t size, bus_size_t align, 
    bus_addr_t maxaddr, bus_dma_tag_t *dmat, bus_dmamap_t *dmam,
    bus_dma_segment_t *segs, void **kvap, bus_addr_t *paddr)
{
	int rseg, ret, phase = 0;

	if (maxaddr) {
		if ((ret = bus_dmatag_subregion(dev->pa.pa_dmat, 0, 
		    maxaddr, dmat, BUS_DMA_NOWAIT) != 0) {
		DRM_DEBUG("dmatag_subregion = %d\n", ret);
		goto bad;
	 	} 
	} else {
		dmat = &(dev->pa.pa_dmat);
	}

	if ((ret = bus_dmamem_alloc(*dmat, size, align, 0, seg, 
	    (sizeof(segs[0]) / sizeof(*segs)), &rseg, BUS_DMA_NOWAIT)) != 0) {
		DRM_DEBUG("dmamem_alloc = %d\n", ret);
		goto bad;
	}
	phase++;

	if ((ret = bus_dmamem_map(*dmat, segs, rseg, size, kvap, 
	    BUS_DMA_NOWAIT | BUS_DMA_COHERENT)) != 0) {
		DRM_DEBUG("dmamem_map = %d\n", ret);
		goto bad;
	}
	phase++;

	if ((ret = bus_dmamap_create(*dmat, size, size / rseg, rseg, 0, 
	    BUS_DMA_NOWAIT, dmam)) != 0) {
		DRM_DEBUG("dmam_create = %d\n", ret);
		goto bad;
	}
	phase++;

	if ((ret = bus_dmamap_load(*dmat, *dmam, *kvap, size, NULL, 
	    BUS_DMA_NOWAIT)) != 0) {
		DRM_DEBUG("dmamap_load = %d\n", ret);
		goto bad;
	}

	*paddr = (*dmam)->dm_segs[0].ds_addr;
	memset(*kvap, 0, size);
	return (0);

 bad:
	if (phase > 2)
		bus_dmamap_destroy(dmat, *dmam);
	if (phase > 1)
		bus_dmamem_unmap(dmat, *kvap, size);
	if (phase > 0)
		bus_dmamem_free(dmat, segs, 1);
	/*
	 * bus_dmatag_destroy counts references, so don't worry about
	 * untoward tag desctruction. 
	 */
		 bus_dmatag_destroy(dmat);

	return (ret);
}

void
drm_dmamem_xfree(bus_dma_tag_t dmat, bus_size_t size, bus_dmamap_t dmam,
    bus_dma_segment_t *segs, void *kvap)
{

	bus_dmamap_unload(dmat, dmam);
	bus_dmamap_destroy(dmat, dmam);
	bus_dmamem_unmap(dmat, kvap, size);
	bus_dmamem_free(dmat, segs, 1);
	bus_dmatag_destroy(dmat);
}
